@@ -17,6 +17,10 @@ import calendar
 import binascii
 import random
 
+# as pytz is optional in thirdparty libs but we need it for good support under
+# python2, just test that it's well installed
+import pytz  # noqa
+
 try:
     from collections import OrderedDict
 except ImportError:
@@ -40,10 +44,12 @@ special_dow_re = re.compile(
     (r'^(?P<pre>((?P<he>(({WEEKDAYS})(-({WEEKDAYS}))?)').format(WEEKDAYS=WEEKDAYS) +
     (r'|(({MONTHS})(-({MONTHS}))?)|\w+)#)|l)(?P<last>\d+)$').format(MONTHS=MONTHS)
 )
+re_star = re.compile('[*]')
 hash_expression_re = re.compile(
     r'^(?P<hash_type>h|r)(\((?P<range_begin>\d+)-(?P<range_end>\d+)\))?(\/(?P<divisor>\d+))?$'
 )
 VALID_LEN_EXPRESSION = [5, 6]
+EXPRESSIONS = {}
 
 
 def timedelta_to_seconds(td):
@@ -151,9 +157,13 @@ class croniter(object):
 
     def __init__(self, expr_format, start_time=None, ret_type=float,
                  day_or=True, max_years_between_matches=None, is_prev=False,
-                 hash_id=None):
+                 hash_id=None, implement_cron_bug=False, second_at_beginning=None,
+                 expand_from_start_time=False):
         self._ret_type = ret_type
         self._day_or = day_or
+        self._implement_cron_bug = implement_cron_bug
+        self.second_at_beginning = bool(second_at_beginning)
+        self._expand_from_start_time = expand_from_start_time
 
         if hash_id:
             if not isinstance(hash_id, (bytes, str)):
@@ -177,7 +187,13 @@ class croniter(object):
         self.cur = None
         self.set_current(start_time, force=False)
 
-        self.expanded, self.nth_weekday_of_month = self.expand(expr_format, hash_id=hash_id)
+        self.expanded, self.nth_weekday_of_month = self.expand(
+            expr_format,
+            hash_id=hash_id,
+            from_timestamp=self.dst_start_time if self._expand_from_start_time else None,
+            second_at_beginning=second_at_beginning
+        )
+        self.expressions = EXPRESSIONS[(expr_format, hash_id, second_at_beginning)]
         self._is_prev = is_prev
 
     @classmethod
@@ -189,6 +205,8 @@ class croniter(object):
                 "[{0}] is not acceptable".format(" ".join(expressions)))
 
     def get_next(self, ret_type=None, start_time=None):
+        if start_time and self._expand_from_start_time:
+            raise ValueError("start_time is not supported when using expand_from_start_time = True.")
         self.set_current(start_time, force=True)
         return self._get_next(ret_type or self._ret_type, is_prev=False)
 
@@ -223,7 +241,7 @@ class croniter(object):
         """
         Converts a UNIX timestamp `timestamp` into a `datetime` object.
         """
-        result = datetime.datetime.utcfromtimestamp(timestamp)
+        result = datetime.datetime.fromtimestamp(timestamp, tz=tzutc()).replace(tzinfo=None)
         if self.tzinfo:
             result = result.replace(tzinfo=tzutc()).astimezone(self.tzinfo)
 
@@ -254,19 +272,30 @@ class croniter(object):
                             "is acceptable.")
 
         # exception to support day of month and day of week as defined in cron
+        dom_dow_exception_processed = False
         if (expanded[2][0] != '*' and expanded[4][0] != '*') and self._day_or:
-            bak = expanded[4]
-            expanded[4] = ['*']
-            t1 = self._calc(self.cur, expanded, nth_weekday_of_month, is_prev)
-            expanded[4] = bak
-            expanded[2] = ['*']
-
-            t2 = self._calc(self.cur, expanded, nth_weekday_of_month, is_prev)
-            if not is_prev:
-                result = t1 if t1 < t2 else t2
+            # If requested, handle a bug in vixie cron/ISC cron where day_of_month and day_of_week form
+            # an intersection (AND) instead of a union (OR) if either field is an asterisk or starts with an asterisk
+            # (https://crontab.guru/cron-bug.html)
+            if self._implement_cron_bug and (re_star.match(self.expressions[2]) or re_star.match(self.expressions[4])):
+                # To produce a schedule identical to the cron bug, we'll bypass the code that
+                # makes a union of DOM and DOW, and instead skip to the code that does an intersect instead
+                pass
             else:
-                result = t1 if t1 > t2 else t2
-        else:
+                bak = expanded[4]
+                expanded[4] = ['*']
+                t1 = self._calc(self.cur, expanded, nth_weekday_of_month, is_prev)
+                expanded[4] = bak
+                expanded[2] = ['*']
+
+                t2 = self._calc(self.cur, expanded, nth_weekday_of_month, is_prev)
+                if not is_prev:
+                    result = t1 if t1 < t2 else t2
+                else:
+                    result = t1 if t1 > t2 else t2
+                dom_dow_exception_processed = True
+
+        if not dom_dow_exception_processed:
             result = self._calc(self.cur, expanded,
                                 nth_weekday_of_month, is_prev)
 
@@ -378,6 +407,8 @@ class croniter(object):
                     if is_prev:
                         d += relativedelta(months=diff_month)
                         reset_day = DAYS[d.month - 1]
+                        if d.month == 2 and self.is_leap(d.year) is True:
+                            reset_day += 1
                         d += relativedelta(
                             day=reset_day, hour=23, minute=59, second=59)
                     else:
@@ -611,7 +642,7 @@ class croniter(object):
             return False
 
     @classmethod
-    def _expand(cls, expr_format, hash_id=None):
+    def _expand(cls, expr_format, hash_id=None, second_at_beginning=False, from_timestamp=None):
         # Split the expression in components, and normalize L -> l, MON -> mon,
         # etc. Keep expr_format untouched so we can use it in the exception
         # messages.
@@ -637,12 +668,28 @@ class croniter(object):
         if len(expressions) not in VALID_LEN_EXPRESSION:
             raise CroniterBadCronError(cls.bad_length)
 
+        if len(expressions) == 6 and second_at_beginning:
+            # move second to last to process by same logical
+            expressions.append(expressions.pop(0))
+
         expanded = []
         nth_weekday_of_month = {}
 
         for i, expr in enumerate(expressions):
             for expanderid, expander in EXPANDERS.items():
-                expr = expander(cls).expand(efl, i, expr, hash_id=hash_id)
+                expr = expander(cls).expand(efl, i, expr, hash_id=hash_id, from_timestamp=from_timestamp)
+
+            if "?" in expr:
+                if expr != "?":
+                    raise CroniterBadCronError(
+                        "[{0}] is not acceptable.  Question mark can not "
+                        "used with other characters".format(expr_format))
+                if i not in [2, 4]:
+                    raise CroniterBadCronError(
+                        "[{0}] is not acceptable.  Question mark can only used "
+                        "in day_of_month or day_of_week".format(expr_format))
+                # currently just trade `?` as `*`
+                expr = "*"
 
             e_list = expr.split(',')
             res = []
@@ -717,6 +764,10 @@ class croniter(object):
                     ):
                         raise CroniterBadCronError(
                             "{0} is out of bands".format(expr_format))
+
+                    if from_timestamp:
+                        low = cls._get_low_from_current_date_number(i, int(step), int(from_timestamp))
+
                     try:
                         rng = range(low, high + 1, step)
                     except ValueError as exc:
@@ -766,7 +817,14 @@ class croniter(object):
             res = set(res)
             res = sorted(res, key=lambda i: "{:02}".format(i) if isinstance(i, int) else i)
             if len(res) == cls.LEN_MEANS_ALL[i]:
-                res = ['*']
+                # Make sure the wildcard is used in the correct way (avoid over-optimization)
+                if (
+                    (i == 2 and '*' not in expressions[4]) or
+                    (i == 4 and '*' not in expressions[2])
+                ):
+                    pass
+                else:
+                    res = ['*']
 
             expanded.append(['*'] if (len(res) == 1
                                       and res[0] == '*')
@@ -777,18 +835,22 @@ class croniter(object):
             dow_expanded_set = set(expanded[4])
             dow_expanded_set = dow_expanded_set.difference(nth_weekday_of_month.keys())
             dow_expanded_set.discard("*")
-            if dow_expanded_set:
+            # Skip: if it's all weeks instead of wildcard
+            if dow_expanded_set and len(set(expanded[4])) != cls.LEN_MEANS_ALL[4]:
                 raise CroniterUnsupportedSyntaxError(
                     "day-of-week field does not support mixing literal values and nth day of week syntax.  "
                     "Cron: '{}'    dow={} vs nth={}".format(expr_format, dow_expanded_set, nth_weekday_of_month))
 
+        EXPRESSIONS[(expr_format, hash_id, second_at_beginning)] = expressions
         return expanded, nth_weekday_of_month
 
     @classmethod
-    def expand(cls, expr_format, hash_id=None):
+    def expand(cls, expr_format, hash_id=None, second_at_beginning=False, from_timestamp=None):
         """Shallow non Croniter ValueError inside a nice CroniterBadCronError"""
         try:
-            return cls._expand(expr_format, hash_id=hash_id)
+            return cls._expand(expr_format, hash_id=hash_id,
+                               second_at_beginning=second_at_beginning,
+                               from_timestamp=from_timestamp)
         except (ValueError,) as exc:
             error_type, error_instance, traceback = sys.exc_info()
             if isinstance(exc, CroniterError):
@@ -801,28 +863,60 @@ class croniter(object):
                 raise CroniterBadCronError("{0}".format(exc))
 
     @classmethod
-    def is_valid(cls, expression, hash_id=None):
+    def _get_low_from_current_date_number(cls, i, step, from_timestamp):
+        dt = datetime.datetime.fromtimestamp(from_timestamp, tz=datetime.timezone.utc)
+        if i == 0:
+            return dt.minute % step
+        if i == 1:
+            return dt.hour % step
+        if i == 2:
+            return ((dt.day - 1) % step) + 1
+        if i == 3:
+            return dt.month % step
+        if i == 4:
+            return (dt.weekday() + 1) % step
+
+        raise ValueError("Can't get current date number for index larger than 4")
+
+    @classmethod
+    def is_valid(cls, expression, hash_id=None, encoding='UTF-8',
+                 second_at_beginning=False):
+        if hash_id:
+            if not isinstance(hash_id, (bytes, str)):
+                raise TypeError('hash_id must be bytes or UTF-8 string')
+            if not isinstance(hash_id, bytes):
+                hash_id = hash_id.encode(encoding)
         try:
-            cls.expand(expression, hash_id=hash_id)
+            cls.expand(expression, hash_id=hash_id,
+                       second_at_beginning=second_at_beginning)
         except CroniterError:
             return False
         else:
             return True
 
     @classmethod
-    def match(cls, cron_expression, testdate, day_or=True):
-        cron = cls(cron_expression, testdate, ret_type=datetime.datetime, day_or=day_or)
+    def match(cls, cron_expression, testdate, day_or=True, second_at_beginning=False):
+        return cls.match_range(cron_expression, testdate, testdate, day_or, second_at_beginning)
+
+    @classmethod
+    def match_range(cls, cron_expression, from_datetime, to_datetime,
+                    day_or=True, second_at_beginning=False):
+        cron = cls(cron_expression, to_datetime, ret_type=datetime.datetime,
+                   day_or=day_or, second_at_beginning=second_at_beginning)
         td, ms1 = cron.get_current(datetime.datetime), relativedelta(microseconds=1)
         if not td.microsecond:
             td = td + ms1
         cron.set_current(td, force=True)
         tdp, tdt = cron.get_current(), cron.get_prev()
         precision_in_seconds = 1 if len(cron.expanded) == 6 else 60
-        return (max(tdp, tdt) - min(tdp, tdt)).total_seconds() < precision_in_seconds
+        duration_in_second = (to_datetime - from_datetime).total_seconds() + precision_in_seconds
+        return (max(tdp, tdt) - min(tdp, tdt)).total_seconds() < duration_in_second
 
 
 def croniter_range(start, stop, expr_format, ret_type=None, day_or=True, exclude_ends=False,
-                   _croniter=None):
+                   _croniter=None,
+                   second_at_beginning=False,
+                   expand_from_start_time=False):
     """
     Generator that provides all times from start to stop matching the given cron expression.
     If the cron expression matches either 'start' and/or 'stop', those times will be returned as
@@ -835,7 +929,7 @@ def croniter_range(start, stop, expr_format, ret_type=None, day_or=True, exclude
     auto_rt = datetime.datetime
     # type is used in first if branch for perfs reasons
     if (
-        type(start) != type(stop) and not (
+        type(start) is not type(stop) and not (
             isinstance(start, type(stop)) or
             isinstance(stop, type(start)))
     ):
@@ -843,7 +937,7 @@ def croniter_range(start, stop, expr_format, ret_type=None, day_or=True, exclude
             "The start and stop must be same type.  {0} != {1}".
             format(type(start), type(stop)))
     if isinstance(start, (float, int)):
-        start, stop = (datetime.datetime.utcfromtimestamp(t) for t in (start, stop))
+        start, stop = (datetime.datetime.fromtimestamp(t, tzutc()).replace(tzinfo=None) for t in (start, stop))
         auto_rt = float
     if ret_type is None:
         ret_type = auto_rt
@@ -857,7 +951,9 @@ def croniter_range(start, stop, expr_format, ret_type=None, day_or=True, exclude
             stop -= ms1
     year_span = math.floor(abs(stop.year - start.year)) + 1
     ic = _croniter(expr_format, start, ret_type=datetime.datetime, day_or=day_or,
-                   max_years_between_matches=year_span)
+                   max_years_between_matches=year_span,
+                   second_at_beginning=second_at_beginning,
+                   expand_from_start_time=expand_from_start_time)
     # define a continue (cont) condition function and step function for the main while loop
     if start < stop:        # Forward
         def cont(v):
