@@ -133,6 +133,14 @@ class CroniterTest(base.TestCase):
         self.assertEqual(n4.day, 29)
         self.assertEqual(n4.month, 2)
 
+    def testDay2(self):
+        base3 = datetime(2024, 2, 28)
+        itr2 = croniter('* * 29 2 *', base3)
+        n3 = itr2.get_prev(datetime)
+        self.assertEqual(n3.year, 2020)
+        self.assertEqual(n3.month, 2)
+        self.assertEqual(n3.day, 29)
+
     def testWeekDay(self):
         base = datetime(2010, 2, 25)
         itr = croniter('0 0 * * sat', base)
@@ -202,6 +210,23 @@ class CroniterTest(base.TestCase):
         self.assertEqual(n3.day, 1)
         self.assertEqual(n3.year, 2010)
 
+    def testDomDowVixieCronBug(self):
+        expr = '0 16 */2 * sat'
+
+        # UNION OF "every odd-numbered day" and "every Saturday"
+        itr = croniter(expr, start_time=datetime(2023, 5, 2), ret_type=datetime)
+        self.assertEqual(itr.get_next(), datetime(2023, 5, 3, 16, 0, 0))    # Wed May 3 2023
+        self.assertEqual(itr.get_next(), datetime(2023, 5, 5, 16, 0, 0))    # Fri May 5 2023
+        self.assertEqual(itr.get_next(), datetime(2023, 5, 6, 16, 0, 0))    # Sat May 6 2023
+        self.assertEqual(itr.get_next(), datetime(2023, 5, 7, 16, 0, 0))    # Sun May 7 2023
+
+        # INTERSECTION OF "every odd-numbered day" and "every Saturday"
+        itr = croniter(expr, start_time=datetime(2023, 5, 2), ret_type=datetime, implement_cron_bug=True)
+        self.assertEqual(itr.get_next(), datetime(2023, 5, 13, 16, 0, 0))    # Sat May  13 2023
+        self.assertEqual(itr.get_next(), datetime(2023, 5, 27, 16, 0, 0))    # Sat May  27 2023
+        self.assertEqual(itr.get_next(), datetime(2023, 6, 3, 16, 0, 0))     # Sat June  3 2023
+        self.assertEqual(itr.get_next(), datetime(2023, 6, 17, 16, 0, 0))    # Sun June 17 2023
+
     def testMonth(self):
         base = datetime(2010, 1, 25)
         itr = croniter('0 0 1 * *', base)
@@ -319,11 +344,15 @@ class CroniterTest(base.TestCase):
         # Test each field individually
         self.assertEqual(croniter('0-59 0 1 1 0').expanded[m], wildcard)
         self.assertEqual(croniter('0 0-23 1 1 0').expanded[h], wildcard)
-        self.assertEqual(croniter('0 0 1-31 1 0').expanded[d], wildcard)
+        self.assertEqual(croniter('0 0 1-31 1 0').expanded[dow], [0])
+        self.assertEqual(croniter('0 0 1-31 1 *').expanded[d], wildcard)
         self.assertEqual(croniter('0 0 1 1-12 0').expanded[mon], wildcard)
-        self.assertEqual(croniter('0 0 1 1 0-6').expanded[dow], wildcard)
-        self.assertEqual(croniter('0 0 1 1 1-7').expanded[dow], wildcard)
-        self.assertEqual(croniter('0 0 1 1 1-7,sat#3').expanded[dow], wildcard)
+        self.assertEqual(croniter('0 0 1 1 0-6').expanded[dow], [0, 1, 2, 3, 4, 5, 6])
+        self.assertEqual(croniter('0 0 * 1 0-6').expanded[dow], wildcard)
+        self.assertEqual(croniter('0 0 1 1 1-7').expanded[dow], [0, 1, 2, 3, 4, 5, 6])
+        self.assertEqual(croniter('0 0 1 1 1-7,sat#3').expanded[dow], [0, 1, 2, 3, 4, 5, 6])
+        self.assertEqual(croniter('0 0 * 1 1-7').expanded[dow], wildcard)
+        self.assertEqual(croniter('0 0 * 1 1-7,sat#3').expanded[dow], wildcard)
         self.assertEqual(croniter('0 0 1 1 0 0-59').expanded[s], wildcard)
         # Real life examples
         self.assertEqual(croniter('30 1-12,0,10-23 15-21 * fri').expanded[h], wildcard)
@@ -871,6 +900,7 @@ class CroniterTest(base.TestCase):
         self.assertTrue(croniter.is_valid('0 * * * *'))
         self.assertFalse(croniter.is_valid('0 * *'))
         self.assertFalse(croniter.is_valid('* * * janu-jun *'))
+        self.assertTrue(croniter.is_valid('H 0 * * *', hash_id='abc'))
 
     def test_exactly_the_same_minute(self):
         base = datetime(2018, 3, 5, 12, 30, 50)
@@ -940,7 +970,7 @@ class CroniterTest(base.TestCase):
         dt = datetime(2019, 1, 14, 0, 0, 0, 0)
         for i in range(10):
             c = croniter("0 0 * * 2-4 *", start_time=dt)
-            dt = datetime.utcfromtimestamp(c.get_next())
+            dt = datetime.fromtimestamp(c.get_next(), dateutil.tz.tzutc()).replace(tzinfo=None)
             ret.append(dt)
             dt += timedelta(days=1)
         sret = ["{0}".format(r) for r in ret]
@@ -960,7 +990,7 @@ class CroniterTest(base.TestCase):
         dt = datetime(2019, 1, 14, 0, 0, 0, 0)
         for i in range(10):
             c = croniter("0 0 * * 1-7 *", start_time=dt)
-            dt = datetime.utcfromtimestamp(c.get_next())
+            dt = datetime.fromtimestamp(c.get_next(), dateutil.tz.tzutc()).replace(tzinfo=None)
             ret.append(dt)
             dt += timedelta(days=1)
         sret = ["{0}".format(r) for r in ret]
@@ -983,7 +1013,7 @@ class CroniterTest(base.TestCase):
         for i in range(10):
             # c = croniter("0 0 * * Mon-Sun *", start_time=dt)
             c = croniter("0 0 * * Wed-Sun *", start_time=dt)
-            dt = datetime.utcfromtimestamp(c.get_next())
+            dt = datetime.fromtimestamp(c.get_next(), tz=dateutil.tz.tzutc()).replace(tzinfo=None)
             ret.append(dt)
             dt += timedelta(days=1)
         sret = ["{0}".format(r) for r in ret]
@@ -1057,6 +1087,79 @@ class CroniterTest(base.TestCase):
             day_or=False
         ))
 
+    def test_match_range(self):
+        self.assertTrue(croniter.match_range(
+            "0 0 * * *",
+            datetime(2019, 1, 13, 0, 59, 0, 0),
+            datetime(2019, 1, 14, 0, 1, 0, 0)
+        ))
+        self.assertFalse(croniter.match_range(
+            "0 0 * * *",
+            datetime(2019, 1, 13, 0, 1, 0, 0),
+            datetime(2019, 1, 13, 0, 59, 0, 0)
+        ))
+        self.assertTrue(croniter.match_range(
+            "0 0 * * * 1",
+            datetime(2023, 5, 25, 0, 0, 0, 0),
+            datetime(2023, 5, 25, 0, 0, 2, 0)
+        ))
+        self.assertFalse(croniter.match_range(
+            "0 0 * * * 1",
+            datetime(2023, 5, 25, 0, 0, 2, 0),
+            datetime(2023, 5, 25, 0, 0, 4, 0)
+        ))
+        self.assertTrue(croniter.match_range(
+            "0 0 * * * 1",
+            datetime(2023, 5, 25, 0, 0, 1, 0),
+            datetime(2023, 5, 25, 0, 0, 4, 0)
+        ))
+        self.assertTrue(croniter.match_range(
+            "31 * * * *",
+            datetime(2019, 1, 14, 1, 30, 0, 0),
+            datetime(2019, 1, 14, 1, 31, 0, 0)
+        ))
+        self.assertTrue(croniter.match_range(
+            "0 0 10 * wed",
+            datetime(2020, 6, 9, 0, 0, 0, 0),
+            datetime(2020, 6, 11, 0, 0, 0, 0),
+            day_or=True
+        ))
+        self.assertTrue(croniter.match_range(
+            "0 0 10 * fri",
+            datetime(2020, 6, 10, 0, 0, 0, 0),
+            datetime(2020, 6, 11, 0, 0, 0, 0),
+            day_or=True
+        ))
+        self.assertTrue(croniter.match_range(
+            "0 0 10 * fri",
+            datetime(2020, 6, 11, 0, 0, 0, 0),
+            datetime(2020, 6, 12, 0, 0, 0, 0),
+            day_or=True
+        ))
+        self.assertTrue(croniter.match_range(
+            "0 0 10 * wed",
+            datetime(2020, 6, 9, 0, 0, 0, 0),
+            datetime(2020, 6, 12, 0, 0, 0, 0),
+            day_or=False
+        ))
+        self.assertFalse(croniter.match_range(
+            "0 0 10 * fri",
+            datetime(2020, 6, 8, 0, 0, 0, 0),
+            datetime(2020, 6, 9, 0, 0, 0, 0),
+            day_or=False
+        ))
+        self.assertFalse(croniter.match_range(
+            "0 0 10 * fri",
+            datetime(2020, 6, 7, 0, 0, 0, 0),
+            datetime(2020, 6, 11, 0, 0, 0, 0),
+            day_or=False
+        ))
+        self.assertFalse(croniter.match_range(
+            "2 4 1 * wed",
+            datetime(2019, 1, 1, 3, 2, 0, 0),
+            datetime(2019, 1, 1, 5, 2, 0, 0),
+            day_or=False
+        ))
 
     def test_dst_issue90_st31ny(self):
         tz = pytz.timezone("Europe/Paris")
@@ -1341,6 +1444,17 @@ class CroniterTest(base.TestCase):
         self.assertListEqual(getn(cron_b, 3), expect_b)
         self.assertListEqual(getn(cron_c, 5), expect_c)
 
+    def test_configure_second_location(self):
+        base = datetime(2010, 8, 25, 0)
+        itr = croniter('59 58 1 * * *', base, second_at_beginning=True)
+        n = itr.get_next(datetime)
+        self.assertEqual(n.year, base.year)
+        self.assertEqual(n.month, base.month)
+        self.assertEqual(n.day, base.day)
+        self.assertEqual(n.hour, 1)
+        self.assertEqual(n.minute, 58)
+        self.assertEqual(n.second, 59)
+
     def test_nth_out_of_range(self):
         with self.assertRaises(CroniterBadCronError):
             croniter("0 0 * * 1#7")
@@ -1353,6 +1467,21 @@ class CroniterTest(base.TestCase):
         with self.assertRaises(CroniterBadCronError):
             croniter("0 0 * * L8")
 
+    def test_question_mark(self):
+        base = datetime(2010, 8, 25, 15, 56)
+        itr = croniter('0 0 1 * ?', base)
+        n = itr.get_next(datetime)
+        self.assertEqual(n.year, base.year)
+        self.assertEqual(n.month, 9)
+        self.assertEqual(n.day, 1)
+        self.assertEqual(n.hour, 0)
+        self.assertEqual(n.minute, 0)
+
+    def test_invalid_question_mark(self):
+        self.assertRaises(CroniterBadCronError, croniter, "? * * * *")
+        self.assertRaises(CroniterBadCronError, croniter, "* ? * * *")
+        self.assertRaises(CroniterBadCronError, croniter, "* * ?,* * *")
+
     def test_issue_47(self):
         base = datetime(2021, 3, 30, 4, 0)
         itr = croniter('0 6 30 3 *', base)
@@ -1548,5 +1677,84 @@ class CroniterTest(base.TestCase):
         self.assertEqual(ret, rets)
         croniter.expand("30 6 1-7 MAY MON#1")
 
+    def test_bug_62_leap(self):
+        ret = croniter("15 22 29 2 *", datetime(2024, 2, 29)).get_prev(datetime)
+        self.assertEqual(ret, datetime(2020, 2, 29, 22, 15))
+
+    def test_expand_from_start_time_minute(self):
+        seven_seconds_interval_pattern = "*/7 * * * *"
+        ret1 = croniter(seven_seconds_interval_pattern, start_time=datetime(2024, 7, 11, 10, 11), expand_from_start_time=True).get_next(datetime)
+        self.assertEqual(ret1, datetime(2024, 7, 11, 10, 18))
+
+        ret2 = croniter(seven_seconds_interval_pattern, start_time=datetime(2024, 7, 11, 10, 12), expand_from_start_time=True).get_next(datetime)
+        self.assertEqual(ret2, datetime(2024, 7, 11, 10, 19))
+
+        ret3 = croniter(seven_seconds_interval_pattern, start_time=datetime(2024, 7, 11, 10, 11), expand_from_start_time=True).get_prev(datetime)
+        self.assertEqual(ret3, datetime(2024, 7, 11, 10, 4))
+
+        ret4 = croniter(seven_seconds_interval_pattern, start_time=datetime(2024, 7, 11, 10, 12), expand_from_start_time=True).get_prev(datetime)
+        self.assertEqual(ret4, datetime(2024, 7, 11, 10, 5))
+
+    def test_expand_from_start_time_hour(self):
+        seven_hours_interval_pattern = "0 */7 * * *"
+        ret1 = croniter(seven_hours_interval_pattern, start_time=datetime(2024, 7, 11, 15, 0), expand_from_start_time=True).get_next(datetime)
+        self.assertEqual(ret1, datetime(2024, 7, 11, 22, 0))
+
+        ret2 = croniter(seven_hours_interval_pattern, start_time=datetime(2024, 7, 11, 16, 0), expand_from_start_time=True).get_next(datetime)
+        self.assertEqual(ret2, datetime(2024, 7, 11, 23, 0))
+
+        ret3 = croniter(seven_hours_interval_pattern, start_time=datetime(2024, 7, 11, 15, 0), expand_from_start_time=True).get_prev(datetime)
+        self.assertEqual(ret3, datetime(2024, 7, 11, 8, 0))
+
+        ret4 = croniter(seven_hours_interval_pattern, start_time=datetime(2024, 7, 11, 16, 0), expand_from_start_time=True).get_prev(datetime)
+        self.assertEqual(ret4, datetime(2024, 7, 11, 9, 0))
+
+    def test_expand_from_start_time_date(self):
+        five_days_interval_pattern = "0 0 */5 * *"
+        ret1 = croniter(five_days_interval_pattern, start_time=datetime(2024, 7, 12), expand_from_start_time=True).get_next(datetime)
+        self.assertEqual(ret1, datetime(2024, 7, 17))
+
+        ret2 = croniter(five_days_interval_pattern, start_time=datetime(2024, 7, 13), expand_from_start_time=True).get_next(datetime)
+        self.assertEqual(ret2, datetime(2024, 7, 18))
+
+        ret3 = croniter(five_days_interval_pattern, start_time=datetime(2024, 7, 12), expand_from_start_time=True).get_prev(datetime)
+        self.assertEqual(ret3, datetime(2024, 7, 7))
+
+        ret4 = croniter(five_days_interval_pattern, start_time=datetime(2024, 7, 13), expand_from_start_time=True).get_prev(datetime)
+        self.assertEqual(ret4, datetime(2024, 7, 8))
+
+    def test_expand_from_start_time_month(self):
+        three_monts_interval_pattern = "0 0 1 */3 *"
+        ret1 = croniter(three_monts_interval_pattern, start_time=datetime(2024, 7, 1), expand_from_start_time=True).get_next(datetime)
+        self.assertEqual(ret1, datetime(2024, 10, 1))
+
+        ret2 = croniter(three_monts_interval_pattern, start_time=datetime(2024, 8, 1), expand_from_start_time=True).get_next(datetime)
+        self.assertEqual(ret2, datetime(2024, 11, 1))
+
+        ret3 = croniter(three_monts_interval_pattern, start_time=datetime(2024, 7, 1), expand_from_start_time=True).get_prev(datetime)
+        self.assertEqual(ret3, datetime(2024, 4, 1))
+
+        ret4 = croniter(three_monts_interval_pattern, start_time=datetime(2024, 8, 1), expand_from_start_time=True).get_prev(datetime)
+        self.assertEqual(ret4, datetime(2024, 5, 1))
+
+    def test_expand_from_start_time_day_of_week(self):
+        three_monts_interval_pattern = "0 0 * * */2"
+        ret1 = croniter(three_monts_interval_pattern, start_time=datetime(2024, 7, 10), expand_from_start_time=True).get_next(datetime)
+        self.assertEqual(ret1, datetime(2024, 7, 12))
+
+        ret2 = croniter(three_monts_interval_pattern, start_time=datetime(2024, 7, 11), expand_from_start_time=True).get_next(datetime)
+        self.assertEqual(ret2, datetime(2024, 7, 13))
+
+        ret3 = croniter(three_monts_interval_pattern, start_time=datetime(2024, 7, 10), expand_from_start_time=True).get_prev(datetime)
+        self.assertEqual(ret3, datetime(2024, 7, 8))
+
+        ret4 = croniter(three_monts_interval_pattern, start_time=datetime(2024, 7, 11), expand_from_start_time=True).get_prev(datetime)
+        self.assertEqual(ret4, datetime(2024, 7, 9))
+
+    def test_get_next_fails_with_expand_from_start_time_true(self):
+        expanded_croniter = croniter("0 0 */5 * *", expand_from_start_time=True)
+        self.assertRaises(ValueError, expanded_croniter.get_next, datetime, start_time=datetime(2024, 7, 12))
+
+
 if __name__ == '__main__':
     unittest.main()
